# -*- coding: utf-8 -*-
"""
Created on Sun Oct 23 20:12:38 2022

@author: lidon
"""

__all__=['vectorADT','vector','alg','AVL_tree','bin_node','bin_tree',
         'bst','graph','hash_table','linked_list','listADT','queues','stack']


# -*- coding: utf-8 -*-
"""
Created on Thu Oct  6 21:49:38 2022

@author: lidon
"""

from enum import Enum
import math
from tinylib.vector import*
from tinylib.queues import*
from tinylib.stack import*
import random

class v_status(Enum):
    UNDISCOVERED=1
    DISCOVERED=2
    VISITED=3

class e_status(Enum):
    UNDETERMINED=0
    TREE=1
    CROSS=2
    FORWARD=3
    BACKWARD=4


class vertex:
    def __init__(self,data=None,in_deg=0,out_deg=0,status=None):
        self.data=data
        self.in_deg=in_deg
        self.out_deg=out_deg
        self.status=status
        if not self.status:
            self.status=v_status.UNDISCOVERED
        # time stamp
        self.d_time=-1
        self.f_time=-1
        # used in traverse tree
        self.parent=-1
        self.priority=math.inf
    def __repr__(self):
        return str(self.data)

class edge:
    def __init__(self,data=None,weight=1,status=None):
        self.data=data
        self.weight=weight
        self.status=status
        if not status:
            self.status=e_status.UNDETERMINED
    
    def copy(self):
        return edge(self.data,self.weight,self.status)

# e set and v set should be vector
# v set are vertex in graph
# e set is the adj matrix!
class graph:
    def __init__(self,v_set=vector([]),e_set=vector([])):
        self.v_set=v_set
        self.e_set=e_set
        # n and e are vertex num and edge num
        self.n=len(v_set)
        self.e=0
        if self.e_set and len(v_set)>1:
            for i in e_set:
                for j in i:
                    if j:
                        self.e=self.e+1
        
        # handle the case when have v set but no e set
        if v_set and not e_set:
            self.e_set=vector([vector([None for i in range(0,self.n)]) for j in range(0,self.n)])
    
    # return adj matrix
    def __repr__(self):
        mat=vector([])
        if self.n==1:
            mat.append(0)
            return str(mat)
        else:
            for i in range(0,len(self.e_set)):
                mat.append(vector([]))
                for j in range(0,len(self.e_set)):
                    if self.e_set[i][j]:
                        mat[i].append(self.e_set[i][j].weight)
                    else:
                        mat[i].append(0)
        return str(mat)
                        
                        
    
    
    # look up ith vertex
    def vertex(self,i):
        return self.v_set[i].data
    
    # in degree
    def in_degree(self,i):
        return self.v_set[i].in_deg
    
    # out degree
    def out_degree(self,i):
        return self.v_set[i].out_deg
    
    # if there exists an edge from i to j
    def exists(self,i,j):
        return i>=0 and i<self.n and j>=0 and self.e_set[i][j]
    
    # if there exists edge from i to j/from j to i
    def adjacent(self,i,j):
        return self.e_set[i][j] or self.e_set[j][i]
    
    # insert a vertex into the graph
    # return its index
    def insert_vertex(self,vertex):
        for j in range(0,self.n):
            self.e_set[j].append(None)
        self.n=self.n+1
        self.e_set.append(vector([None for k in range(0,self.n)]))
        return self.v_set.append(vertex)
    
    # remove ith vertex
    def remove_vertex(self,i):
        # delete out deg
        for j in range(0,self.n):
            if self.exists(i,j):
                self.v_set[j].in_deg=self.v_set[j].in_deg-1
        # remove the ith row of the matrix
        self.e_set.remove(i)
        # reduce the vertex number
        self.n=self.n-1
        # delete in deg
        for j in range(0,self.n):
            if self.e_set[j][i]:
                self.v_set[j].out_deg=self.v_set[j].out_deg-1
            self.e_set[j].remove(i)
        # return the deleted vertex
        return self.v_set.remove(i)
    
    # query the status of edge between i and j
    # otherwise return None (if no edge)
    def e_status(self,i,j):
        if e_set[i][j]:
            return e_set[i][j].status
        return None
    
    # query the data of edge between i and j
    # return None if no edge
    def e_data(self,i,j):
        if e_set[i][j]:
            return e_set[i][j].data
        return None
    
    # return weight of edge between i and j
    def weight(self,i,j):
        return e_set[i][j].weight
    
    # insert an edge
    # will automatically cover the old edge (if exists any)!
    def insert_edge(self,edge,i,j):
        self.e_set[i][j]=edge
        self.e=self.e+1
        self.v_set[i].out_deg=self.v_set[i].out_deg+1
        self.v_set[j].in_deg=self.v_set[j].in_deg+1
    
    # remove an edge
    # return the rmoved edge
    def remove_edge(self,i,j):
        removed=self.e_set[i][j].copy()
        self.e_set[i][j]=None
        self.e=self.e-1
        self.v_set[i].out_deg=self.v_set[i].out_deg-1
        self.v_set[j].in_deg=self.v_set[j].in_deg-1
        return removed
    
    # return the next neighbor of i wrt j
    # if does not exist, return -1
    def next_neighbor(self,i,j):
        j=j-1
        while j>-1 and not self.exists(i,j):
            j=j-1
        return j
    
    # return next adjacent of i wrt j (ignore direction)
    # if doesn't exists, return -1
    def next_adjacent(self,i,j):
        j=j-1
        while j>-1 and not (self.exists(i,j) or self.exists(j,i)):
            j=j-1
        return j
    
    # return first neighbor of i
    def first_neighbor(self,i):
        return self.next_neighbor(i,self.n)
    
    def first_adjacent(self,i):
        return self.next_adjacent(i,self.n)
    
    def d_time(self,i):
        return self.v_set[i].d_time
    
    def f_time(self,i):
        return self.v_set[i].f_time
    
    def parent(self,i):
        return self.v_set[i].parent
    
    def priority(self,i):
        return self.v_set[i].priority
    
    # reset all vertex/edge status after traversing
    def reset(self):
        for i in range(0,self.n):
            self.v_set[i].status=v_status.UNDISCOVERED
            self.v_set[i].d_time=-1
            self.v_set[i].f_time=-1
            self.v_set[i].parent=-1
            self.v_set[i].priority=-1
            for j in range(0,self.n):
                if self.exists(i,j):
                    self.e_set[i][j].status=e_status.UNDETERMINED
    
    # breadth first search algorithm
    # start from i-th vertex
    # func is the operation to the vertex
    
    # bfs in a sub-branch
    def bfs_branch(self,i,func=print):
        Q=queue()
        self.v_set[i].status=v_status.DISCOVERED
        Q.enqueue(i)
        while Q:
            v=Q.dequeue()
            func(self.v_set[v])
            self.v_set[v].status=v_status.VISITED
            u=self.first_neighbor(v)
            while u>-1:
                if self.v_set[u].status==v_status.UNDISCOVERED:
                    self.v_set[u].status=v_status.DISCOVERED
                    Q.enqueue(u)
                u=self.next_neighbor(v,u)
    
    # bfs through all sub-branches
    def bfs(self,i,func=print):
        self.reset()
        self.bfs_branch(i,func)
        for v in range(0,self.n):
            if self.v_set[v].status==v_status.UNDISCOVERED:
                self.bfs_branch(v,func)
    
    # depth first traverse algorithm
    # start from i-th vertex, apply func() to all visited vertex
    def dfs_branch(self,i,func=print):
        if self.v_set[i].status!=v_status.UNDISCOVERED:
            return
        else:
            func(self.v_set[i])
            self.v_set[i].status=v_status.DISCOVERED
            j=self.first_neighbor(i)
            while j>-1:
                if self.v_set[j].status==v_status.UNDISCOVERED:
                    self.dfs_branch(j,func)
                j=self.next_neighbor(i,j)
            self.v_set[i].status=v_status.VISITED
    
    # dfs through all sub branches
    def dfs(self,i,func=print):
        self.reset()
        self.dfs_branch(i,func)
        for v in range(0,self.n):
            if self.v_set[v].status==v_status.UNDISCOVERED:
                self.dfs_branch(v,func)
            
    
    # dfs for topological sorting
    # kept hidden from user
    def topo_dfs_branch(self,i,s=stack([])):
        if self.v_set[i].status!=v_status.UNDISCOVERED:
            return s
        else:
            self.v_set[i].status=v_status.DISCOVERED
            j=self.first_neighbor(i)
            while j>-1:
                if self.v_set[j].status==v_status.UNDISCOVERED:
                    s=self.topo_dfs_branch(j,s)
                j=self.next_neighbor(i,j)
            self.v_set[i].status=v_status.VISITED
            s.push(i)
            return s
    
    # topological sort
    def t_sort(self,i=0):
        self.reset()
        s=stack([])
        s=self.topo_dfs_branch(i,s)
        for v in range(0,self.n):
            if self.v_set[v].status==v_status.UNDISCOVERED:
                s=self.topo_dfs_branch(v,s)
        
        sort=vector([])
        while s:
            sort.append(s.pop())
        return sort
    
    # priority first search in a branch
    # priority_updater is a function to update the priority of each node
    def pfs_branch(self,i,priority_updater):
        if self.v_set[i].status==v_status.VISITED:
            return
        
        # add the starting point into the pfs tree
        self.v_set[i].priority=0
        self.v_set[i].parent=-1
        self.v_set[i].status=v_status.VISITED
        
        
        while 1:
            w=self.first_neighbor(i)
            while w>-1:
                # enumerate all neighbors, update their priority
                self.priority_updater(i,w)
                w=self.next_neighbor(i,w)
                
            # find point with tallest priority
            shortest=math.inf
            for w in range(0,self.n):
                if self.v_set[w].status==v_status.UNDISCOVERED:
                    if shortest>self.v_set[w].priority:
                        shortest=self.v_set[w].priority
                        i=w
            if self.v_set[i].status==v_status.VISITED:
                break
            self.v_set[i].status=v_status.VISITED
            self.e_set[self.v_set[i].parent][i].status=e_status.TREE
    
    # minimal spanning tree use prim
    # start from i
    # edges in MST will have the status TREE
    # assume undirected graph
    def mst(self,i):
        self.reset()
        new_v=vector([i])
        self.v_set[i].status=v_status.VISITED
        while len(new_v)<self.n:
            start=new_v[0]
            s=new_v[0]
            short=math.inf
            for u in new_v:
                w=self.first_adjacent(u)
                while w>-1:
                    if self.exists(u,w):
                        edg=self.e_set[u][w]
                    else:
                        edg=self.e_set[w][u]
                    if self.v_set[w].status!=v_status.VISITED and edg.weight<short:
                        start=u
                        s=w
                        short=edg.weight
                        #edg.status=e_status.TREE
                    w=self.next_adjacent(u,w)
            self.v_set[s].status=v_status.VISITED
            new_v.append(s)
            if self.exists(start,s):
                self.e_set[start][s].status=e_status.TREE
            else:
                self.e_set[s][start].status=e_status.TREE
    
    # find shortest path tree wrt i'
    # assume undirected graph
    # edges in SPT will have status TREE
    def spt(self,i): 
        self.reset()
        new_v=vector([i])
        self.v_set[i].status=v_status.VISITED
        self.v_set[i].priority=0
        while len(new_v)<self.n:
            start=new_v[0]
            s=new_v[0]
            short=math.inf
            for u in new_v:
                w=self.first_adjacent(u)
                while w>-1:
                    if self.exists(u,w):
                        edg=self.e_set[u][w]
                    else:
                        edg=self.e_set[w][u]
                    if self.v_set[w].status!=v_status.VISITED and self.v_set[u].priority+edg.weight<short:
                        start=u
                        s=w
                        short=self.v_set[start].priority+edg.weight
                        #edg.status=e_status.TREE
                    w=self.next_adjacent(u,w)
            self.v_set[s].status=v_status.VISITED
            
            new_v.append(s)
            if self.exists(start,s):
                self.e_set[start][s].status=e_status.TREE
                edge_weight=self.e_set[start][s].weight
            else:
                self.e_set[s][start].status=e_status.TREE
                edge_weight=self.e_set[s][start].weight
            self.v_set[s].priority=self.v_set[start].priority+edge_weight
    
    # construct a derived graph that only includes edges with TREE status
    # NOTE: this will automatically delete the original edges that 
    # are not members of the tree
    def derive_tree(self):
        #new_v=self.v_set.copy()
        #new_e=self.e_set.copy()
        #new_g=graph(new_v,new_e)
        for i in range(0,self.n):
            for j in range(0,self.n):
                if self.e_set[i][j]:
                    if self.e_set[i][j].status!=e_status.TREE:
                        self.remove_edge(i,j)
                if self.e_set[j][i]:
                    if self.e_set[j][i].status!=e_status.TREE:
                        self.remove_edge(j,i)
                        
    
        
    # priority first search starting from i
    def pfs(self,i,priority_updater):
        self.reset()
        self.pfs_branch(i,priority_updater)
        for i in range(0,self.n):
            self.pfs_branch(i,priority_updater)
    
    
    
    
    # priority updater of prim
    # update v wrt u
    def prim_pu(self,u,v):
        if self.v_set[v].status==v_status.UNDISCOVERED:
            self.v_set[v].priority=self.e_set[u][v].weight
            self.v_set[v].parent=u
    
                        
            
    
        
        
        
            





# -*- coding: utf-8 -*-
"""
Created on Sat Oct  1 21:10:02 2022

@author: lidon
"""

from vector import*
from linked_list import*
from stack import*

# implement some algorithms based on data structures we have implemented


# some additional algorithms implemented by stack
# n the number in base 10, 0<base<16
# base: new base
# return a vector from high to low
# ex. convert(4,2)=[1,0,0]
def convert(n,base):
    # stk is some stack
    digit=[str(i) for i in range(0,9)]+['A','B','C','D','E','F']
    def inner_convert(stk,n,base,digit):
        if n>0:
            stk.push(digit[n%base])
            inner_convert(stk,int(n/base),base,digit)
        return stk
    stk=stack([])
    converted=inner_convert(stk,n,base,digit)
    
    # output the new_stack
    new_stk=stack([])
    while converted.length>0:
        new_stk.push(converted.top())
        converted.pop()
        
    return new_stk

#bracket checking
# check if there the brackets are legal
# brackets may include [],() and {}
def bra_check(string):
    stk=stack([])
    for i in range(0,len(string)):
        if string[i] in ['(','[','{']:
            stk.push(string[i])
        elif string[i]==')':
            if stk.empty():
                return False
            popped=stk.pop()
            if popped!='(':
                return False
        elif string[i]==']':
            if stk.empty():
                return False
            popped=stk.pop()
            if popped!=']':
                return False
        elif string[i]=='}':
            if stk.empty():
                return False
            popped=stk.pop()
            if popped!='}':
                return False
        else:
            a=1
    return stk.empty()

# solve n Queen problem
class queen:
    # x and y: respective coordinates
    def __init__(self,x,y):
        self.x=x
        self.y=y
        
    def __repr__(self):
        return f'({self.x},{self.y})'
        
    # verify if 2 queens are in conflict
    def __eq__(self,other):
        return self.x==other.x or self.y==other.y or (self.x+self.y)==(other.x+other.y) or (self.x-self.y)==(other.x-other.y)
    
    def copy(self):
        return queen(self.x,self.y)

# solve n queens problem
def place_queen(n):
    solution=stack([])
    q=queen(0,0)
    while solution.size()<n:
        # search w.r.t. line, prevent overflow
        if q.y>=n:
            # if overflow, track back and move the last queen to the next block
            q=solution.top()
            solution.pop()
            q.x=q.x+1
        # prevent overflow
        if q.x>=n:
            q.y=q.y+1
            q.x=0
        
        # current queen has contradictio with last queens, let it move again
        if solution.find(q)>=0:
            q.x=q.x+1
        # push the queen into the stack, move the new queen
        else:
            p=q.copy()
            solution.push(p)
            q.x=q.x+1
          
        # prevent overflow
        if q.x>=n:
            q.y=q.y+1
            q.x=0
    return solution

       
        
        
            


# -*- coding: utf-8 -*-
"""
Created on Wed Oct 19 21:39:36 2022

@author: lidon
"""

from bst import*
from vector import*
from enum import Enum

def isprime(x):
    if x<=1:
        return False
    else:
        i=2
        for i in range(2,x):
            if x%i==0:
                return False
        return True

# generate n primes larger than c
def prime_gen(c,n):
    primes=vector([])
    # c even
    if c%2==0:
        i=3
        while len(primes)<n:
            if isprime(c+i):
                primes.append(c+i)
            i=i+2
        return primes
    # c odd
    if c%2==1:
        i=2
        while len(primes)<n:
            if isprime(c+i):
                primes.append(c+i)
            i=i+2
        return primes
        
# enumeration of lazy removal
class entry_status(Enum):
    UNUSED=1
    USED=2

class hash_table:
    # generate a hash table
    # length = c = 5 by default
    # don't use too large p, or it will take too long to 
    # select a prime number
    # p<1048576 is recommended
    def __init__(self,c=5,p=None):
        
        # if the user doesn't specify the prime we use
        if not p:
            primes=prime_gen(c+1000,1)
            p=primes[0]
        # M = length of hash table
        # N= number of queries
        self.M=p
        self.N=0
        # ht is the bucket
        # must consist of entries
        self.ht=vector([None for i in range(0,self.M)])
        self.lazy_removal=vector([entry_status.UNUSED for i in range(0,self.M)])
    
    # the default scheme to generate hash code
    def default_encoder(self,k):
        if type(k)==str:
            h=0
            for i in range(0,len(k)):
                # 37 is a magic number
                h=h+(37**i)*ord(k[i])
            return h%self.M
        else:
            return int(k)%self.M
    
    
    # define hash code
    def hash_code(self,k,encoder=None):
        # if don't specify an encoder, then use the default hash code generator
        if (not encoder):
            return self.default_encoder(k)
        else:
            return encoder(k)
    
    def size(self):
        return self.N
    
    # test if r is lazily removed
    def lazily_removed(self,r):
        if self.lazy_removal[r]==entry_status.USED:
            return True
        return False
    
    def mark_as_removed(self,r):
        self.lazy_removal[r]=entry_status.USED
        
        
    # probe linearly to prevent conflict
    def linear_probe(self,k):
        r=self.hash_code(k)
        while (self.ht[r] and self.ht[r].key!=k) or ((not self.ht[r]) and self.lazily_removed(r)):
            r=(r+1)%self.M
        # by ht[r] and lazy_removal[r] we know if searching is successful
        return r
    
    # get value by hashing
    def get(self,k):
        r=self.linear_probe(k)
        if self.ht[r]:
            return self.ht[r].value
        else:
            # if the key doesn't exist, deny the access
            return False
    
    # remove k
    def remove(self,k):
        r=self.linear_probe(k)
        if not self.ht[r]:
            # doesn't exist any value, removal denied
            return False
        self.ht[r]=None
        self.mark_as_removed(r)
        self.N=self.N-1
        # removal successful
        return True
    
    # find a free bucket
    def probe_for_free(self,k):
        r=self.hash_code(k)
        while self.ht[r]:
            r=(r+1)%self.M
        return r
    
    # insert
    # k: key
    # v: value
    def put(self,k,v):
        r=self.linear_probe(k)
        # can't insert same element
        if self.ht[r]:
            return False
        # find a free bucket
        r=self.probe_for_free(k)
        self.ht[r]=entry(k,v)
        self.N=self.N+1
        # if too many elements, rehash the table
        if 2*self.N>self.M:
            self.rehash()
        # insert successful
        return True
    
    # if too many elements, rehash the whole table
    def rehash(self):
        old_M=self.M
        old_N=self.N
        old_ht=self.ht
        primes=prime_gen(2*old_M,1)
        self.M=primes[0]
        self.N=0
        self.ht=[None for i in range(0,self.M)]
        self.lazy_removal=vector([entry_status.UNUSED for i in range(0,self.M)])
        for r in range(0,len(old_ht)):
            if old_ht[r]:
                self.put(old_ht[r].key,old_ht[r].value)

# implement bucket sort via hash table
# v is the input vector
def bucket_sort(v):
    c=len(v)
    p=prime_gen(c,1)[0]
    
    int_table=hash_table(c,p)
    for i in range(0,len(v)):
        int_table.put(v[i],v[i])
    
    v1=vector([])
    
    for j in range(0,int_table.M):
        if int_table.ht[j]:
            v1.append(int_table.ht[j].value)
    return v1




# -*- coding: utf-8 -*-
"""
Created on Sun Oct  2 21:34:53 2022

@author: lidon
"""
from enum import Enum

class COLOR(Enum):
    RED=1
    BLUE=2

# return height of node p
# if null tree, return -1
def stature(p):
    if p:
        return p.height
    else:
        return -1

# some useful macros
def is_root(x):
    return not x.par
def is_lc(x):
    return (not is_root(x)) and (x==x.par.lc)
def is_rc(x):
    return (not is_root(x)) and (x==x.par.rc)
def has_par(x):
    return not is_root(x)
def has_lc(x):
    return x.lc
def has_rc(x):
    return x.rc
def has_child(x):
    return x.lc or x.rc
# if has both children
def has_children(x):
    return x.lc and x.rc
def is_leaf(x):
    return not x.has_child
# return the sibling of x
def sibling(x):
    if is_lc(x):
        return x.par.rc
    else:
        return x.par.lc
def uncle(x):
    if is_lc(x.par):
        return x.par.par.rc
    else:
        return x.par.par.lc
# return pointer from its parent
def from_par_to(x):
    if is_root(x):
        return x
    else:
        if is_lc(x):
            return x.par.lc
        else:
            return x.par.rc



class bin_node:
    # lc,rc: red children, left children
    # pt: parent
    # npl: null path length
    # color: red or black
    # definition of height: height starts from 0, any single node has height 0, null tree
    # has height 1
    # height of a node is the depth of the subtree starting from this node
    def __init__(self,data=None,par=None,lc=None,rc=None,height=0,npl=1,color=COLOR.RED):
        self.data=data
        self.par=par
        self.lc=lc
        self.rc=rc
        self.height=height
        self.color=color
    
    # return number of all offsprings
    def size(self):
        x=self.lc
        y=self.rc
        # x and y are not null
        if x and y:
            return 1+x.size()+y.size()
        # x not null, y null
        if x and not y:
            return 1+x.size()
        # x null, y not null
        if not x and y:
            return 1+y.size()
        else:
            return 1
    
    # insert as left children, return the new children
    def insert_as_lc(self,e):
        child = bin_node(data=e,par=self)
        self.lc=child
        return child
    
    # insert as right children, return the new children
    def insert_as_rc(self,e):
        child=bin_node(data=e,par=self)
        self.rc=child
        return child
    
    # successor in trave in
    def succ(self):
        s=self
        if self.rc:
            s=self.rc
            while has_lc(s):
                s=s.lc
        else:
            while is_rc(s):
                s=s.par
            s=s.par
        return s
        
    
    # traverse wrt level
    def trav_level(self,func):
        pass
    
    # pre traverse
    def trav_pre(self,func):
        pass
    
    # In traverse
    def trav_in(self,func):
        pass
    
    # post traverse
    def trav_post(self,func):
        pass
    
    
    # reload <=
    def __le__(self,other):
        return self.data<=other.data
    
    # reload ==
    def __eq__(self,other):
        return self.data==other.data
    
  
    
    


# -*- coding: utf-8 -*-
"""
Created on Tue Sep 27 20:37:27 2022

@author: lidon
"""

from listADT import*
import random
from vector import*

# implement find()
def find(self,e,n=None,node=None):
    if n==None and node==None:
        n=self.size
        node=self.trailer
    while n>0:
        node=node.pred
        if e==node.data:
            return node
        n=n-1
    # find failed
    return None
# bind to linked list
linked_list.find=find

# implement insert_as_pred()
# insert e as the pred
def insert_as_pred(self,e):
    new_node=list_node(data=e,pred=self.pred,succ=self)
    self.pred.succ=new_node
    self.pred=new_node
    # return the pointer of the new node
    return self.pred
# bind to list_node
list_node.insert_as_pred=insert_as_pred

# implement insert as succ()
# insert e as the succ
def insert_as_succ(self,e):
    new_node=list_node(data=e,pred=self,succ=self.succ)
    self.succ.pred=new_node
    self.succ=new_node
    # return the pointer of the new node
    return self.succ
# bind to list_node
list_node.insert_as_succ=insert_as_succ

# implement insert_as_first
def insert_as_first(self,e):
    self.size=self.size+1
    return self.header.insert_as_succ(e)
# bind to linked_list
linked_list.insert_as_first=insert_as_first

# implement insert_as_last
def insert_as_last(self,e):
    self.size=self.size+1
    return self.trailer.insert_as_pred(e)
# bind to linked_list
linked_list.insert_as_last=insert_as_last

# implement insert_succ
def insert_succ(self,node,e):
    self.size=self.size+1
    return node.insert_as_succ(e)
# bind to linked list
linked_list.insert_succ=insert_succ

# implement insert_pred
def insert_pred(self,node,e):
    self.size=self.size+1
    return node.insert_as_pred(e)
# bind to linked list
linked_list.insert_pred=insert_pred

# implement copy_nodes()
def copy_nodes(self,node,n):
    new_list=linked_list()
    while n>0:
        new_list.insert_as_last(node.data)
        node=node.succ
        n=n-1
    return new_list
# bind to linked_list
linked_list.copy_nodes=copy_nodes


# implement remove()
def remove(self,node):
    backup=node.data
    node.pred.succ=node.succ
    node.succ.pred=node.pred
    del node
    self.size=self.size-1
    return backup
# bind to linked_list
linked_list.remove=remove

# implement clear()
# remove all nodes apart from header and trailer
# return number of nodes removed
def clear(self):
    old_size=self.size
    while self.size>0:
        self.remove(self.header.succ)
        #self.size=self.size-1
    return old_size
# bind to linked_list
linked_list.clear=clear

# implement deduplicate()
# return number of elements removed
def deduplicate(self):
    old_size=self.size
    # trivial linked list
    if self.size<2:
        return None
    # start from 1th element
    p=self.header.succ.succ
    index=1
    while p.succ!=None:
        if self.find(p.data,index,p):
            tmp=p.succ
            self.remove(p)
            p=tmp
        else:
            index=index+1
            p=p.succ
    return old_size-self.size
# bind to linked_list
linked_list.deduplicate=deduplicate 

# implement traverse
def traverse(self,func):
    p=self.header.succ
    while p.succ!=None:
        func(p.data)
        p=p.succ
# bind to linked_list
linked_list.traverse=traverse

# implement uniquify
# return number of elements removed
def uniquify(self):
    old_size=self.size
    p=self.header.succ
    q=p
    while p.succ!=self.trailer:
        q=p.succ
        if p.data!=q.data:
            p=q
        else:
            self.remove(q)
            q=p.succ
    return old_size-self.size
# bind to linked list
linked_list.uniquify=uniquify

# implement search
# note it returns node <=e with largest rank
def search(self,e,n=None,node=None):
    if n==None and node==None:
        n=self.size
        node=self.trailer
    p=node.pred
    while n>0:
        if p.data<=e:
            return p
        else:
            p=p.pred
            n=n-1
    return None
# bind to linked list
linked_list.search=search

#implement insertion sort
# complexity: O(n^2)
def insertion_sort(self,node=None,n=None):
    if node==None and n==None:
        node=self.header.succ
        n=self.size
    p=node.succ
    for r in range(1,n):
        if self.search(p.data,r,p)==None:
            self.insert_pred(node,p.data)
            p=p.succ
            self.remove(p.pred)
        else:
            self.insert_succ(self.search(p.data,r,p),p.data)
            p=p.succ
            self.remove(p.pred)
# bind to linked list
linked_list.insertion_sort=insertion_sort

# implement selection sort
# complexity: O(n^2)
def selection_sort(self,node=None,n=None):
    if node==None and n==None:
        node=self.header.succ
        n=self.size
        
    p=node
    tmp_head=p.pred
    tmp_tail=p
    for i in range(0,n):
        tmp_tail=tmp_tail.succ
    while n>1:
        maximum=self.select_max(tmp_head.succ,n)
        self.insert_pred(tmp_tail,self.remove(maximum))
        tmp_tail=tmp_tail.pred
        n=n-1
# bind to linked list
linked_list.selection_sort=selection_sort

# implement merge()
def merge(self,p,n,q,m):
    
    
    pp=p.pred
    qq=q.pred
    while m>0:
        if n>0 and p.data<=q.data:
            p=p.succ
            if q==p:
                break
            n=n-1
        else:
            m=m-1
            # can't change the relative position of p and q! (i.e. can't add new pred of p and q)
            # otherwise, the upper level p and q will also be changed
            p.data,q.data=swap(p.data,q.data)
            tmp=q.data
            q=q.succ
            self.remove(q.pred)
            self.insert_succ(p,tmp)
            p=p.succ
            
    p=pp.succ
# bind to linked list
linked_list.merge=merge


# implement merge sort
def merge_sort(self,node=None,n=None):
    if node==None and n==None:
        node=self.header.succ
        n=self.size
    if n<2:
        return 
    p=node
    m=int(n/2)
    q=p
    for i in range(0,m):
        q=q.succ
    
    self.merge_sort(p,m)
    self.merge_sort(q,n-m)
    self.merge(p,m,q,n-m)
# bind to linked list
linked_list.merge_sort=merge_sort



    


# -*- coding: utf-8 -*-
"""
Created on Mon Oct 17 21:36:48 2022

@author: lidon
"""
# implementation of AVL tree
from bst import*

# some useful APIs
def balanced(v):
    return stature(v.lc)==stature(v.rc)
def balance_factor(v):
    return stature(v.lc)-stature(v.rc)
def AVL_balanced(v):
    return (balance_factor(v)>-2) and (balance_factor(v)<2)
# return taller child of v
def taller_child(v):
    if stature(v.lc)>stature(v.rc):
        return v.lc
    if stature(v.lc)<stature(v.rc):
        return v.rc
    # if same height, return the child that is at the same side with its father
    if is_lc(v):
        return v.lc
    else:
        return v.rc


class AVL_tree(bst):
    # v are nodes and t are sub trees
    # order them in in order: t0 v0 t1 v1 t2 v2 t3
    def connect_34(self,v0,v1,v2,t0,t1,t2,t3):
        v0.lc=t0
        if t0:
            t0.par=v0
        v0.rc=t1
        if t1:
            t1.par=v1
        self.update_height(v0)
        v2.lc=t2
        v2.rc=t3
        if t2:
            t2.par=v2
        if t3:
            t3.par=v2
        self.update_height(v2)
        v1.lc=v0
        v1.rc=v2
        v0.par=v1
        v2.par=v1
        self.update_height(v1)
        return v1
    
    # rotate at node v use connect34    
    def rotate(self,v):
        p=v.par
        g=p.var
        # zig
        if is_lc(p):
            # zig zig
            if is_lc(v):
                p.par=g.par
                if g.par:
                    if p.data<g.par.data:
                        g.par.lc=p
                    else:
                        g.par.rc=p
                return self.connect_34(v,p,g,v.lc,v.rc,p.rc,g.rc)
            # zig zag
            if is_rc(v):
                v.par=g.par
                if g.par:
                    if v.data<g.par.data:
                        g.par.lc=v
                    else:
                        g.par.rc=v
                return self.connect_34(p,v,g,p.lc,v.lc,v.rc,g.rc)
        # zag
        if is_rc(p):
            # zag zag
            if is_rc(v):
                p.par=g.par
                if g.par:
                    if p.data<g.par.data:
                        g.par.lc=p
                    else:
                        g.par.rc=p
                return self.connect34(g,p,v,g.lc,p.lc,v.lc,v.rc)
            # zag zig
            if is_lc(v):
                v.par=g.par
                if g.par:
                    if v.data<g.par.data:
                        g.par.lc=v
                    else:
                        g.par.rc=v
                return self.connect34(g,v,p,g.lc,v.lc,v.rc,p.rc)
    
    # insert e into the tree, and automatically balance the tree
    def insert(self,e):
        v,hot=self.search_in(self.root,e,hot=None)
        # make sure the value doesn't exist
        if v:
            return v
        
        # insert the node
        new_v=bin_node(data=e)
        new_v.par=hot
        if hot:
            if e<hot.data:
                hot.lc=new_v
            else:
                hot.rc=new_v
        self.size=self.size+1
        #self.update_height_above(hot)
        # start from v's parent, check if each of its ancestor is unbalanced
        g=hot
        while g:
            # if g unbalanced, rebalance it by rotating
            if not AVL_balanced(g):
                if is_root(g):
                    g=self.rotate(taller_child(taller_child(g)))
                else:
                    if is_lc(g):
                        g.par.lc=self.rotate(taller_child(taller_child(g)))
                    else:
                        g.par.rc=self.rotate(taller_child(taller_child(g)))
                # if g balanced, all its ancestors will be balanced
                break
            else:
                self.update_height(g)
            g=g.par
        
        return new_v
    
    # return true if success, false otherwise
    def remove(self,e):
        v,hot=self.search_in(self.root,e,hot=None)
        # if v doesn't exist
        if not v:
            return False
        self.remove_at(v,hot)
        self.size=self.size-1
        g=hot
        # from hot, check one by one to above and rebalance by rotating
        while g:
            if not AVL_balanced(g):
                if is_root(g):
                    g=self.rotate(taller_child(taller_child(g)))
                else:
                    if is_lc(g):
                        g.par.lc=self.rotate(taller_child(taller_child(g)))
                    else:
                        g.par.rc=self.rotate(taller_child(taller_child(g)))
                # unbalance will propagate in deletion case, so continue searching
            self.update_height(g)
            g=g.par
        # removal successful
        return True
            
        
            


# -*- coding: utf-8 -*-
"""
Created on Sat Oct  1 19:18:30 2022

@author: lidon
"""

from vector import*
from vectorADT import*

class stack(vector):
    def __init__(self,elements):
        super(stack,self).__init__(elements)
    
    # push e to the end of the stack
    def push(self,e):
        self.insert(self.size(),e)
    
    # pop e from the end of the stack
    # return the poped value
    def pop(self):
        return self.remove(self.size()-1)
    
    # return the top element
    def top(self):
        return self.elements[self.size()-1]



